/*
 *  602html: T602 document file to HTML converter
 *
 *  Copyright (c) 2000 Martin Hinner <mhi@penguin.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "config.h"
#include <stdio.h>
#include <getopt.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdlib.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

extern char *version_string;			/* from version.c */
extern unsigned char kamenic_conv[0x100];
extern unsigned char latin2_conv[0x100];
extern unsigned char koi8cs_conv[0x100];

char *helptext = "Usage: %s [OPTIONS] [input-file] [output-file]\n"
"\n"
"Options:\n"
"  -b, --br          End *each* line with <br>\n"
"  -h, --help        Show this message\n"
"  -v, --version     Show version\n"
"\n"
"Report bugs to <mhi@penguin.cz>\n";
char *htmlheader1 = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\n"
"<HTML>\n"
"<HEAD><TITLE></TITLE>\n"
"<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=iso-8859-2\">\n"
"<META NAME=\"GENERATOR\" CONTENT=\"602html (";
char *htmlheader2 = ")\">\n"
"</HEAD>\n"
"<BODY bgcolor=white text=black link=blue alink=blue vlink=blue>\n";
char *htmlfooter = "</BODY>\n"
"</HTML>\n";


static struct option long_options[] =
{				/* For parsing long arguments */
  {"br", no_argument, 0, 'b'},
  {"version", no_argument, 0, 'v'},
  {"help", no_argument, 0, 'h'},
  {0, 0, 0, 0},
};
int option_index;		/* Unused */

#define BUFFER_LEN	20480	/* Input buffer */

int boldface;
int italic;
int underline;
int sup;
int sub;
int wide;
int tall;
int big;
int lastbrk;
char *conv;
int allbreaks;

int
process_line (int outf, unsigned char *line)
{
  int brk;
  int i;
  int len;
  int prevspace;
  int onlyspaces;

  if (!line)
    return -1;

  switch (line[strlen (line) - 1])
    {
    case 0x8d:
      line[strlen (line) - 1] = 0;
      brk = 0;
      break;
    case 0x0d:
      line[strlen (line) - 1] = 0;
      brk = 1;
      break;
    case 0x1A:
      line[strlen (line) - 1] = 0;
      brk = 1;
      break;
    default:
      brk = 0;
    }

  if (line[0] == '@')
    {
      if (!memcmp (line, "@PA", 3))
	write (outf, "<hr>\n", 5);
      else if (!memcmp (line, "@CT ", 4)) {
       switch(line[4]) {
        case '0':
          conv = kamenic_conv;
          break;
        case '1':
          conv = latin2_conv;
          break;
        case '2':
          conv = koi8cs_conv;
          break;
        default:
          fprintf(stderr,"Unknown codepage ignored (%s)\n", line+4);
          break;
       }
      } else
	{
	  write (outf, "<!-- ", 5);
	  write (outf, line, strlen (line));
	  write (outf, " -->\n", 5);
	}
      lastbrk = 1;
    }
  else
    {
      len = strlen (line);

      prevspace = 1;
      onlyspaces = 1;
      for (i = 0; i < len; i++)
	{
          line[i] = conv[line[i]];
	  switch (line[i])
	    {
	    case 0x02:		/* Boldface */
	      if (boldface)
		write (outf, "</b>", 4);
	      else
		write (outf, "<b>", 3);
	      boldface = boldface ? 0 : 1;
	      break;
	    case 0x04:		/* Italic */
	      if (italic)
		write (outf, "</i>", 4);
	      else
		write (outf, "<i>", 3);
	      italic = italic ? 0 : 1;
	      break;
	    case 0x0F:		/* Wide */
	      if (wide)
		write (outf, "</font>", 7);
	      else
		write (outf, "<font size=+1>", 14);
	      wide = wide ? 0 : 1;
	      break;
	    case 0x10:		/* Tall */
	      if (tall)
		write (outf, "</font>", 7);
	      else
		write (outf, "<font size=+2>", 14);
	      tall = tall ? 0 : 1;
	      break;
	    case 0x13:		/* Underline */
	      if (underline)
		write (outf, "</u>", 4);
	      else
		write (outf, "<u>", 3);
	      underline = underline ? 0 : 1;
	      break;
	    case 0x14:		/* Superscript */
	      if (sup)
		write (outf, "</sup>", 6);
	      else
		write (outf, "<sup>", 5);
	      sup = sup ? 0 : 1;
	      break;
	    case 0x16:		/* Subscript */
	      if (sub)
		write (outf, "</sub>", 6);
	      else
		write (outf, "<sub>", 5);
	      sub = sub ? 0 : 1;
	      break;
	    case 0x1D:		/* Big */
	      if (big)
		write (outf, "</font>", 7);
	      else
		write (outf, "<font size=+3>", 14);
	      big = big ? 0 : 1;
	      break;
	    case 0x20:		/* Space */
              if ((!lastbrk) && (onlyspaces) )
                 break;
	      if ((!prevspace) && (line[i + 1] != 0x20))
		write (outf, " ", 1);
	      else
		write (outf, "&nbsp;", 6);
	      break;
	    case 0xFE:		/* NB Space */
	      onlyspaces = 0;
	      write (outf, "&nbsp;", 6);
	      break;
	    default:
              onlyspaces = 0;
	      write (outf, &line[i], 1);
	      break;
	    }			/* switch */
	  prevspace = (line[i] == 0x20);
	}			/* for */

      if (brk || allbreaks)
	write (outf, "<br>", 4);
      write (outf, "\n", 1);
      lastbrk = (brk || allbreaks)?1:0;
    }
  return 0;
}

int
convert602 (int inf, int outf, char *inname, char *outname)
{
  char *buffer;
  int nread;
  int i;
  int startline, len;

  buffer = (char *) malloc (BUFFER_LEN);
  if (!buffer)
    {
      fprintf (stderr, "%s: Not enough memory\n", inname);
      return -1;
    }

  nread = read (inf, buffer, BUFFER_LEN);
  if (nread < 0)
    {
      perror (inname);
      return -1;
    }

  if (write (outf, htmlheader1, strlen (htmlheader1)) < 0)
    {
      perror (outname);
      return -1;
    }
  if (write (outf, version_string, strlen (version_string)) < 0)
    {
      perror (outname);
      return -1;
    }
  if (write (outf, htmlheader2, strlen (htmlheader2)) < 0)
    {
      perror (outname);
      return -1;
    }


  while (1)
    {
      startline = 0;
      for (i = 0; i < nread; i++)
	{
	  if (buffer[i] == '\n')
	    {
	      buffer[i] = 0;
	      if (process_line (outf, (unsigned char *) buffer + startline) < 0)
		return -3;
	      startline = i + 1;
	    }
	}
      len = nread - startline;
      memcpy (buffer, buffer + startline, len);

      nread = read (inf, buffer + len, BUFFER_LEN - len);
      if (nread < 0)
	{
	  perror (outname);
	  return -1;
	}
      if (nread == 0)
	{
	  nread += len;
	  break;
	}
      nread += len;
    }

  buffer[nread] = 0;
  if (process_line (outf, (unsigned char *) buffer) < 0)
    return -3;

  if (write (outf, htmlfooter, strlen (htmlfooter)) < 0)
    {
      perror (outname);
      return -1;
    }

  return 0;
}

int
main (int argc, char **argv)
{
  char c;
  char *argv0;
  int inf, outf;
  int rc;

  allbreaks = 0;

  argv0 = argv[0];
  /* First we parse all the command line arguments */
  while ((c = getopt_long (argc, argv, "acvh",
			   long_options, &option_index)) != -1)
    {
      switch (c)
	{
        case 'b':
          allbreaks = 1;
          break;
	case 'v':
	  fprintf (stderr, "602html version %s, built " __DATE__ " on "
		BUILDHOST "\n",
		version_string);
	  exit (0);
	case 'h':
	  fprintf (stderr, helptext, argv0);
	  exit (0);
	case '?':
	  fprintf (stderr, helptext, argv0);
	  exit (1);
	}
    }

  argc -= optind;
  argv += optind;

  if (argc != 2)
    {
      fprintf (stderr, "%s: Missing file arguments\n"
	       "Try '%s --help' for more information\n", argv0, argv0);
      exit (1);
    }

  if (!strcmp (argv[0], "-"))
    {
      inf = STDIN_FILENO;
    }
  else
    {
      inf = open (argv[0], O_RDONLY);
      if (inf < 0)
	{
	  perror (argv[0]);
	  exit (1);
	}
    }

  if (!strcmp (argv[1], "-"))
    {
      outf = STDOUT_FILENO;
    }
  else
    {
      outf = open (argv[1], O_RDWR | O_CREAT | O_TRUNC, 0644);
      if (outf < 0)
	{
	  perror (argv[1]);
	  exit (1);
	}
    }

  boldface = 0;
  italic = 0;
  lastbrk = 1;
  sub = 0;
  sup = 0;
  wide = 0;
  tall = 0;
  big = 0;
  underline = 0;
  conv = kamenic_conv;

  rc = convert602 (inf, outf, inf == STDIN_FILENO ? "STDIN" : argv[0],
		   outf == STDIN_FILENO ? "STDOUT" : argv[1]);

  close (inf);
  close (outf);

  /* Solaris bug */
  exit (rc < 0 ? 2 : 0);
  return 0;
}
